// --- [ TWI_Multi_Master.c ] -------------------------------------------------
//
//        tab = 3 - 13.03.2012

/*****************************************************************************
*
* File              : TWI_Multi_Master.c
*
* AppNote           : AVR311 / 315
*
*
* nderungen        : Die Zuweisung der TWI-Adresse erfolgt direkt ohne Umweg
*							 TWI_bufPtr:
*							 Eine berprfung der Pointer beim Senden und Empfangen findet statt,
*           	       berlauf ber die Array-Grenze wird verhindert
*						    Abfrage von TWI_statusReg ber Funktionen (nicht mehr als globale Variable)
*                     Die Namen einiger Funktionen sind zur Unterscheidung der Master-Bibliothek gendert
* 							 in ISR(TWI) bei switch / default: -> Stop senden !!
*
****************************************************************************/

#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "TWI_Multi_Master.h"
#include "TWI_State_Codes.h"
#include "global.h"

#define TRUE            1
#define FALSE           0

#define TWI_SLAVE_RX    0x01                      // die States fr den TWI_Multi_Master_Mode
#define TWI_SLAVE_TX    0x02
#define TWI_SLAVE       0x04
#define SLAVE_MASK      0x07

#define TWI_MASTER_RX   0x10
#define TWI_MASTER_TX   0x20
#define TWI_MASTER      0x40
#define MASTER_MASK     0x70

#define TWI_PORT_PIN    PINC                       // an diesen Pins wird auf low getestet
#define SDA_PIN         PC4
#define SCL_PIN         PC5


uint8_t SLA_rx_buf[SLA_RX_BUF_SIZE];               // Die Puffer des Slave getrennt fr Senden und Empfangen
uint8_t SLA_tx_buf[SLA_TX_BUF_SIZE];               //
uint8_t MA_buf[MA_BUF_SIZE];                       // Nach Arbitrierungsfehler bleiben die Sendedaten erhalten
                                                   // und knnen bei einem neuen Sendeversuch verschickt werden
volatile uint8_t SLA_Rx_Idx = 0;			            // enthlt die Anzahl der vom Slave empfangenen Bytes
volatile uint8_t MA_msg_size;                      // Number of bytes to be transmitted.

volatile uint8_t TWI_Arbitration_State = 0;			// Flag, das bei Arbitrierungsfehler mit Inhalt von TSWR beschrieben wird

volatile uint8_t TWI_success;
volatile uint8_t TWI_rx_complete = FALSE;

#if DEBUG == TRUE
   uint8_t t_buf[SLA_RX_BUF_SIZE * 2];
   uint8_t t_idx;
#endif

volatile uint8_t TWI_Rep_Start = FALSE;            // Es soll anstelle eines STOP ein REPEATED START gesendet werden
volatile uint8_t TWI_Multi_Master_Mode = 0;        // Der Modus, in dem das Modul arbeitet: als Slave (default) oder als Master

// --- lokale Funktionen des MultiMasters -------------------------------------

uint8_t TWI_MA_read( uint8_t *msg, uint8_t msgSize );
uint8_t TWI_MA_write( uint8_t *msg, uint8_t msgSize );
uint8_t TWI_MA_write_read(uint8_t *, uint8_t out, uint8_t in);
void    TWI_Check_Bus(void);

// ----------------------------------------------------------------------------
// Grundlegende Initialisierung fr Master und Slave Modus
void TWI_MM_Init( void )
{
   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
	TWAR = TWI_SLA_ADR;                          // Set own TWI slave address.
	TWDR = 0xFF;                                 // Default content = SDA released.
	TWCR = (1<<TWEN)|                          	// Enable TWI-interface and release TWI pins.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // Do ACK on any requests.
          (0<<TWWC);
}

/*
// ist als Makro definiert
// ----------------------------------------------------------------------------
// Liefert den TWSR-Code der letzten bertragung
uint8_t TWI_Get_State(void)
{
//while (TWI_BUSY);
//return(TWI_state);
return(TWSR);
}
*/

#if DEBUG == TRUE
// -----------------------------------------------------------------------
// lscht die letzten States und schreibt den Header mit Loggeradresse und
// eigener TWI-SLAVE-Adr
void TWI_Clear_States(void)
{
   uint8_t i;
   t_buf[0] = 4;                 // TWI-Adresse des Interface
   t_buf[1] = TWI_SLA_ADR;       // die eigene Adresse voranstellen

   for(i = 2; i < (SLA_RX_BUF_SIZE * 2) ; i++) t_buf[i] = 0;
   t_idx = 2;
}

// --------------------------------------------------------------------------
// gibt die aufgezeichneten TWSR-States auf dem Logger aus
void TWI_Get_States(void)
{
   TWI_MM_Write(&t_buf[0], (SLA_RX_BUF_SIZE * 2));
}
#endif

// ----------------------------------------------------------------------------
// Prft ber einen hinreichend langen Zeitraum, ob SCL low ist.
// Wenn ja, dann wird die Wartezeit soweit verlngert, bis der Bus
// frei ist.
// Bei einem Repeated Start ist diese Prfung aber berflssig und
// sogar kontraproduktiv: SCL ist hier low bis zum folgenden TWI-START
//
// Wenn im Multi-Master-Modus

void TWI_Check_Bus(void)
{
   uint8_t t = 10;
   if (!(TWCR & (1<<TWIE))) return;       // bei einem Repeated Start beenden

   while(t--)
   {
      _delay_us(1);
      if (!(TWI_PORT_PIN & (1<<SCL_PIN))) t = 10;
   }
}

// --- MULTI_MASTER - Funktionen --------------------------------------------------------
// Setzt die globale Variable TWI_Rep_Start.
// Muss vor der Transaktion ausgefhrt werden, die mit einem REPEATED START abgeschlossen werden soll
// Dannach muss es unbedingt wieder gelscht werden.
// Wenn nicht bleiben SDA und SCL auf low.

// Genaugenommen wird dabei der TWINT nicht zurckgesetzt, SCL bleibt low,
// der Bus wird nicht freigegeben, weil das STOP ausgleibt.
// Das folgende TWI-Start wird dadurch als Repeated Start erkannt.

// Parameter: TWI_REPEATED_START (1) oder TWI_STOP (0)
void TWI_MM_Set_Stop_Mode(uint8_t m)
{
   TWI_Rep_Start = m;
}

// ---------------------------------------------------------------------------------------
// Ruft die lokale Funktion TWI_MA_write() auf.
// Bei einem Fehler wird die TWI_MA_write() unverzglich beendet.
//
// Danach wird in einer Warteschleife gewartet, whrend die TWI-Schnittstelle passiv ist
// und alle Anfragen mit NACK beantwortet.
//
// Die Anzahl der neuen Versuche wird mit der Konstanten ERROR_RETRYS festgelegt.
// Die Dauer der Wartezeit setzt sich zusammen aus einem konstanten Anteil ( ERROR_WAIT_CYCLES ),
// zu der die TWI_SLA_ADR/2 addiert wird.
// Damit entsteht fr jeden Master eine andere Wartezeit.
// Die Wartezeit muss in einem vernnftigen Verhltnis zur mittleren Datenlnge/bertragungsdauer stehen.
//
// Die Funktion liefert im Erfolgsfalle TRUE zurck, sonst FALSE

uint8_t TWI_MM_Write(uint8_t * data, uint8_t cnt)
{
   uint8_t  i = ERROR_RETRYS;
   uint8_t  e = 0, j;

   while(i--)
   {
      if ((e = TWI_MA_write(data, cnt))) break;
      j = ERROR_WAIT_CYCLES + (TWI_SLA_ADR / 2);
      while(j--) _delay_us(WAIT_US);
   }
   if(e == FALSE)                            // im Fehlerfall den Normalzustand herstellen
   {
      TWI_Rep_Start = FALSE;
      TWCR = (1<<TWEN | 1<<TWIE | 1<<TWEA | 0<<TWINT);
   }
   return(e);
}

// ---------------------------------------------------------------------------------------
// Ruft die lokale Funktion TWI_MA_read() auf.
// Bei einem Fehler wird die TWI_MA_read() unverzglich beendet.
//
// Danach wird in einer Warteschleife gewartet, whrend die TWI-Schnittstelle passiv ist
// und alle Anfragen mit NACK beantwortet.
//
// Die Anzahl der neuen Versuche wird mit der Konstanten ERROR_RETRYS festgelegt.
// Die Dauer der Wartezeit setzt sich zusammen aus einem konstanten Anteil ( ERROR_WAIT_CYCLES ),
// zu der die TWI_SLA_ADR/2 addiert wird.
// Damit entsteht fr jeden Master eine andere Wartezeit.
// Die Wartezeit muss in einem vernnftigen Verhltnis zur mittleren Datenlnge/bertragungsdauer stehen.
//
// Die Funktion liefert im Erfolgsfalle TRUE zurck, sonst FALSE

uint8_t TWI_MM_Read(uint8_t * data, uint8_t cnt)
{
   uint8_t  i = ERROR_RETRYS;
   uint8_t  e = 0, j;

   while(i--)
   {
      if ((e = TWI_MA_read(data, cnt))) break;
      j = ERROR_WAIT_CYCLES + (TWI_SLA_ADR / 2);
      while(j--) _delay_us(WAIT_US);
   }
   if (e == FALSE)                              // im Fehlerfall den Normalzustand herstellen
   {
      TWI_Rep_Start = FALSE;
      TWCR = (1<<TWEN | 1<<TWIE | 1<<TWEA | 0<<TWINT);
   }
   return(e);
}

// ---------------------------------------------------------------------------------------
// Verbindet zwei Transaktionen mit einem REPEATED START
//
// Zuerst werden Daten geschrieben, danach Daten gelesen.
// Die bergabe der Daten erfolgt mit dem selben Array.
// Ruft die lokale Funktion TWI_MA_write() und TWI_MA_read() auf.
// Bei einem Fehler wird die TWI_MA_write() unverzglich beendet.
//
// Danach wird in einer Warteschleife gewartet, whrend die TWI-Schnittstelle passiv ist
// und alle Anfragen mit NACK beantwortet.
//
// Die Anzahl der neuen Versuche wird mit der Konstanten ERROR_RETRYS festgelegt.
// Die Dauer der Wartezeit setzt sich zusammen aus einem konstanten Anteil ( ERROR_WAIT_CYCLES ),
// zu der die TWI_SLA_ADR/2 addiert wird.
// Damit entsteht fr jeden Master eine andere Wartezeit.
// Die Wartezeit muss in einem vernnftigen Verhltnis zur mittleren Datenlnge/bertragungsdauer stehen.
//
// Die Funktion liefert im Erfolgsfalle TRUE zurck, sonst FALSE


uint8_t TWI_MM_Write_Read(uint8_t * data, uint8_t tx, uint8_t rx)
{
   uint8_t  i = ERROR_RETRYS;
   uint8_t  e = 0, j;


   while(i--)
   {
      TWI_Rep_Start = TRUE;                     // Sendet ein Repeated Start anstelle eines STOP
      e = TWI_MA_write(data, tx);               //
      TWI_Rep_Start = FALSE;
      if (e)                                    // Read nur angehen, wenn Write erfolgreich war
         {
         if ((e = (TWI_MA_read(data, rx)))) break;
         }
      j = ERROR_WAIT_CYCLES + (TWI_SLA_ADR / 2);
      while(j--) _delay_us(WAIT_US);
   }
   return(e);
}

// --- SLAVE-ROUTINES --------------------------------------------------------
//
// ----------------------------------------------------------------------------
// Gibt die Anzahl der empfangenen Bytes zurck,
// Rckgabe : 0  - sind keine Daten empfangen worden
// Rckgabe : 1  - es ist genau 1 Byte empfangen - die eigene Adresse + Write-Bit
// Rckgabe : >1 - es sind (n - 1) Datenbyte empfangen, zustzlich die Adresse in Byte.0

uint8_t TWI_SLA_Rx_Cnt(void)
{
if (TWI_rx_complete == FALSE) return(0);
else TWI_rx_complete = FALSE;

return (SLA_Rx_Idx);								      // der Index beeinhaltet die Anzahl der empf. Bytes
}

// ----------------------------------------------------------------------------
// Stellt Daten im lokalen Array TWI_Tx_buf[] zur Abholung durch den Master bereit
void TWI_SLA_Put_Data( uint8_t *msg, uint8_t msgSize )
{
	uint8_t i;

	for (i = 0; i < msgSize; i++)                // Copy data that may be transmitted if the TWI Master requests data.
		SLA_tx_buf[i] = msg[i];
}

// ----------------------------------------------------------------------------
// Kopiert die empfangenen Daten im Array TWI_Rx_buf[] in das Array
// dessen Adresse bergeben wurde

void TWI_SLA_Get_Data( uint8_t *msg)
{
	uint8_t i;
                                                // Copy data from Transceiver buffer.
	for (i = 0; i < SLA_Rx_Idx; i++) msg[i] = SLA_rx_buf[i];
	SLA_Rx_Idx = 0;									   // damit die Daten nur einmal gelesen werden
}

// --- MASTER-ROUTINES -------------------------------------------------------------------
//


// ----------------------------------------------------------------------------------------
// Schreibt die Daten eines Arrays ber die TWI-Schnittstelle raus
// Parameter 1: Ein Pointer auf ein Array
//              Das erste Byte muss die Adresse des TWI-Slaves sein, linksbndig ausgerichtet,
//              das Bit.0 (Read/Write-Bit) ist ohne Bedeutung, es wird automatisch gesetzt
// Parameter 2: Die Anzahl der zu schreibenden Bytes zuzglich '1' fr die Adresse in Byte 0

uint8_t TWI_MA_write( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

   TWI_Arbitration_State = 0;
   TWI_success = FALSE;

   MA_msg_size = msgSize;                       // Number of data to transmit.
   MA_buf[0]  = (msg[0] & 0xFE);          	   // Store slave address with Write-Bit setting.
                                                // Die Daten umkopieren nach TWI_MA_buf[]
   for ( i = 1; i < msgSize; i++ ) MA_buf[i] = msg[i];

   TWI_Check_Bus();

   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
         (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag.
         (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|       // Initiate a START condition.
         (0<<TWWC);

   while ( TWI_BUSY );                          // Wait until TWI is ready for next transmission.
                                                // In der while() Schleife wurden ggf. gendert:
                                                // - TWI_success

   // --- Es sind Fehler aufgetreten ------------------------------------------
   if(!(TWI_success)) PORTB |= (1<<PB0);  else PORTB &= ~(1<<PB0);

   // Fr ein Repeated Start darf das TWI-Interface erst mit dem folgenden TWI-Start verndert werden
   // SCL bleibt auf low
   // Fr alle anderen Flle wird das Interface mit einem Stop wieder fr den Empfang aktiviert

   // Wenn kein Repeated Start aktiviert ist und auch bei einem Fehler wird ein STOP gesendet!
   if ((TWI_Rep_Start == FALSE) || (TWI_success == FALSE))
   {
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
        (1<<TWIE)|(1<<TWINT)|              	   // Enable TWI Interupt and clear the flag to send byte
        (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	   // Send ACK
        (0<<TWWC);
   }

   return(TWI_success);
}

// ----------------------------------------------------------------------------------------
// Schreibt die Daten eines Arrays ber die TWI-Schnittstelle raus
// Parameter 1: Ein Pointer auf ein Array
//              Das erste Byte muss die Adresse des TWI-Slaves sein, linksbndig ausgerichtet,
//              das Bit.0 (Read/Write-Bit) ist ohne Bedeutung, es wird automatisch gesetzt
// Parameter 2: Die Anzahl der zu lesenden Bytes zuzglich '1' fr die Adresse in Byte 0
//
// Das Byte.0 ist die TWI-Adresse, die Daten beginnen bei Byte.1 !!!
//

uint8_t TWI_MA_read( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

//   TWI_Multi_Master_Mode = TWI_MASTER;       // wichtig in der ISR
   TWI_Arbitration_State = 0;
   TWI_success = FALSE;

   MA_msg_size = msgSize;                    // Number of data to transmit.
   MA_buf[0] = (msg[0] | 0x01);              // Store slave address with Read-Bit setting.

   TWI_Check_Bus();

   TWCR = (1<<TWEN)|                         // TWI Interface enabled.
         (1<<TWIE)|(1<<TWINT)|               // Enable TWI Interupt and clear the flag.
         (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|    // Initiate a START condition.
         (0<<TWWC);

// Jetzt werden die Daten im Hintergrund abgeholt,
// wenn die bertragung abgeschlossen ist, das wird zuerst
// TWIE gelscht und gleichzeitig  TWSTO gesetzt - dadurch wird ein TWI-Stop generiert
// ist das TWI-Stop ausgefhrt, wird auf TWSTO gelscht
// Bei einer erfolgreichen Aktion wurde TWI_success auf TRUE gesetzt.
// Bei einem Arbitrierungsfehler ist TWI_Aritration_State == TWSR > 0

   while ( TWI_BUSY );                       // Wait until TWI is ready for next transmission.

   // --- Es sind Fehler aufgetreten ------------------------------------------
   if(!(TWI_success)) PORTB |= (1<<PB0); else PORTB &= ~(1<<PB0);

   if (TWI_success)            					   // Last transmission competed successfully.
   {                                            // Die Daten wurden in der ISR in TWI_MA_buf[] eingelesen
      for ( i = 0; i < msgSize; i++ )           // Copy data from Transceiver buffer.
      {
         msg[i] = MA_buf[i];
      }
   }

   // Fr ein Repeated Start darf das TWI-Interface erst mit dem folgenden TWI-Start verndert werden
   // SCL bleibt auf low
   // Fr alle anderen Flle wird das Interface mit einem Stop wieder fr den Empfang aktiviert

   // Wenn kein Repeated Start aktiviert ist und auch bei einem Fehler wird ein STOP gesendet!
   if ((TWI_Rep_Start == FALSE) || (TWI_success == FALSE))
   {
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
        (1<<TWIE)|(1<<TWINT)|              	   // Enable TWI Interupt and clear the flag to send byte
        (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	   // Send ACK
        (0<<TWWC);
   }

   return( TWI_success );
}

// ---- ISR -------------------------------------------------------------------

ISR(TWI_vect)
{
   static uint8_t idx = 0;

#if DEBUG == TRUE
   t_buf[t_idx] = TWSR;
   if (t_idx < (SLA_RX_BUF_SIZE * 2 - 1)) t_idx++;
#endif

   if (TWSR > 0x5F )                                // Das sind dann die Slave Status Codes
   {
      switch (TWSR)
      {
// TWI Slave Transmitter status codes				   // ###############################################

// Der Controller schreibt die Daten aus SLA_Tx_buf[], als Index wird idx benutzt
// Eigene Slave-Adresse + read empfangen und mit ACK quittiert

      case TWI_STX_ADR_ACK:            				// Own SLA+R has been received; ACK has been returned
         idx = 0;
         TWI_Multi_Master_Mode = TWI_SLAVE_TX;    // TX setzen

// der Slave Transmitter hat ein Byte gesendet, der Master mit ACK quittiert
      case TWI_STX_DATA_ACK:           				// Data byte in TWDR has been transmitted; ACK has been received
                                                   // Es werden solange Daten gesendet, bis der Master mit NACK beendet
         TWDR = SLA_tx_buf[idx];                   // Der idx wird auf die Gre des Array beschrnkt
         if (idx < SLA_RX_BUF_SIZE -1) idx++;      // man kann nicht ber die Grenzen eines Arrays hinaus lesen
         break;

// TWI Slave Receiver status codes					   // ###############################################

// der Controller empfngt Daten und legt sie im Arry SLA_rx_buf[] ab
// eigene Slave-Adresse + write empfangen und mit ACK quittiert
      case TWI_SRX_GEN_ACK:            				// General call address has been received; ACK has been returned
      case TWI_SRX_ADR_ACK: 								// Own SLA+W has been received ACK has been returned

         SLA_rx_buf[0] = TWDR;                     // eigene Adresse / General Call ablegen
         SLA_Rx_Idx = 1;      							// Index fr die nachfolgenden Daten auf 0 stellen
         TWI_Multi_Master_Mode = TWI_SLAVE_RX;     // RX setzen
         break;

// Master im write Modus (Slave im read Modus), Byte empfangen und mit ACK quittiert
      case TWI_SRX_GEN_DATA_ACK:       				// Previously addressed with general call;
                                                   // data has been received; ACK has been returned
      case TWI_SRX_ADR_DATA_ACK:       				// Previously addressed with own SLA+W; data has been received;
                                                   // ACK has been returned
                                                   // Daten, die nicht mehr ins Rx_Array passen, werden ignoriert
         if (SLA_Rx_Idx < SLA_RX_BUF_SIZE) SLA_rx_buf[SLA_Rx_Idx++] = TWDR;
         break;

      case TWI_SRX_STOP_RESTART:       				// A STOP condition or repeated START condition has been received
                                                   // while still addressed as Slave
                                                   // Diese Situation ist auf dem MultiMaster zweideutig:
                                                   // ein Repeated Start kann aus dem Transmitter oder Receiver-Modus
                                                   // sich ereignen. Aber nur im Receiver-Modus sind Daten empfangen worden.
         if (TWI_Multi_Master_Mode == TWI_SLAVE_RX)
            {
            TWI_rx_complete = TRUE;                // Flag fr den Empfang neuer Daten
            TWI_Multi_Master_Mode = FALSE;
            }
         break;

// alle anderen Flle laufen ber default: ------------------------------------
// case TWI_STX_DATA_ACK_LAST_BYTE: 				   // Last data byte in TWDR has been transmitted (TWEA = 0);
                                                   // ACK has been received
// case TWI_STX_DATA_NACK:                         // Data byte in TWDR has been transmitted;
                                                   // NOT ACK has been received
// case TWI_SRX_ADR_DATA_NACK:      				   // Previously addressed with own SLA+W; data has been received;
                                                   // NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				   // Previously addressed with general call;
                                                   // data has been received; NOT ACK has been returned
      default:
         break;
      }
      TWCR = (1<<TWEN)|                            // TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag to send byte
            (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|       // Send ACK after next reception
            (0<<TWWC);
	}

   else															// TWSR < 0x5F - das sind die Master Status Codes
   {

      switch (TWSR)
      {
// MASTER ########################################################################################

	// General TWI Master status codes				// ###############################################
// ein TWI-(RE)-Start wurde gesendet, nun muss die Adresse folgen
      case TWI_START:              						// START has been transmitted
      case TWI_REP_START:          						// Repeated START has been transmitted
         idx = 0;			           						// Set buffer pointer to the start location

	// TWI Master Transmitter status codes
// die Adresse + W wurde gesendet und vom Slave mit ACK quittiert
      case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received

// eine Datenbyte wurde gesendet und vom Slave mit ACK quittiert
      case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received

// wenn noch weitere Bytes gesendet werden sollen ...
         if (idx < MA_msg_size )
         {
            TWDR = MA_buf[idx++];					   // Adresse bzw. Daten nach TWDR schreiben
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to send byte
                  (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|	// Send NACK after reception
                  (0<<TWWC);
			}

         else                                      // alle Bytes sind geschrieben, nun beenden
         {                                         // Der Restart wird vom nachfolgenden TWI_MA_write() ausgefhrt
                                                   // hier wird SCL auf low gehalten, da TWINT nicht zurckgesetzt wird
            if (TWI_Rep_Start)                     // Kein STOP bei einem Repeated Start
            {
                TWCR = (1<<TWEN)|                  // TWI Interface enabled
                     (0<<TWIE)|(0<<TWINT)|         // Nichts ndern, SCL stretchen,
                     (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Initiate nothing.
                     (0<<TWWC);
            }

            else                                   // Send STOP after last byte
            {                                      // Set status bits to completed successfully.
               TWCR = (1<<TWEN)|                   // TWI Interface enabled
                     (0<<TWIE)|(1<<TWINT)|         // Disable TWI Interrupt and clear the flag
                     (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
                     (0<<TWWC);
            }
            TWI_success = TRUE;
         }
			break;

// ein Datenbyte wurde empfangen und vom Master mit ACK quittiert
	// TWI Master Receiver status codes				   // ###############################################
      case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
         MA_buf[idx++] = TWDR;						   // gelesene Daten in den Puffer schreiben

// die Adresse + R wurde gesendet und vom Slave mit ACK quittiert
      case TWI_MRX_ADR_ACK:                        // SLA+R has been transmitted and ACK received

// es folgen noch weiter Bytes
         if (idx < (MA_msg_size -1 ))       			// Detect the last byte to NACK it.
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
                  (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
                  (0<<TWWC);
			}
// das folgende zu lesende Byte ist das letzte, danach wird kein ACK (also ein NACK) gesendet
         else                                      // Send NACK after next reception
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
                  (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
                  (0<<TWWC);
         }
         break;

// das letzte Byte wurde empfangen und vom Master mit NACK quittiert
// nun muss der Master ein Stop senden
      case TWI_MRX_DATA_NACK:                      // Data byte has been received and NACK transmitted

         MA_buf[idx] = TWDR;
         TWI_success = TRUE;

         if (TWI_Rep_Start)
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (0<<TWIE)|(0<<TWINT)|            // Disable TWI Interrupt and do not clear the flag
                  (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Initiate no STOP condition.
                  (0<<TWWC);
         }
         else
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
                  (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
                  (0<<TWWC);
         }
         break;

	// General TWI Master status codes				   // ###############################################

	// -------------------------------------------     Arbitration lost:
      case TWI_SRX_GEN_ACK_M_ARB_LOST:             // Arbitration lost in SLA+R/W as Master; General call address has been received;
                                                   // ACK has been returned
      case TWI_SRX_ADR_ACK_M_ARB_LOST:             // Arbitration lost in SLA+R/W as Master; own SLA+W has been received;
                                                   // ACK has been returned
/*
         SLA_rx_buf[0] = TWDR;                     // eigene Adresse / General Call ablegen
         SLA_Rx_Idx = 1;      							// Index fr die nachfolgenden Daten auf 0 stellen

         TWI_Multi_Master_Mode = TWI_SLAVE_RX;

         TWI_Arbitration_State = TWSR;             // TWI_Arbitration_State > 0 = Fehler
         TWCR =(1<<TWEN)|                         	// TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send NACK after reception
               (0<<TWWC);
         break;
*/
      case TWI_STX_ADR_ACK_M_ARB_LOST:             // Arbitration lost in SLA+R/W as Master;
 /*                                                // own SLA+R has been received; ACK has been returned
         idx = 0;
         TWI_Multi_Master_Mode = TWI_SLAVE_TX;

         TWI_Arbitration_State = TWSR;             // TWI_Arbitration_State > 0 = Fehler
         TWCR =(1<<TWEN)|                         	// TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send NACK after reception
               (0<<TWWC);
         break;
*/
      case TWI_ARB_LOST:                           // Arbitration lost
         TWI_Arbitration_State = TWSR;             // TWI_Arbitration_State > 0 = Fehler
/* luft ber default:
         TWCR =(1<<TWEN)|                         	// TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send NACK after reception
               (0<<TWWC);
         break;      */

// der Rest luft ber default ------------------------------------------------
// case TWI_MTX_DATA_NACK:                     	   // Data byte has been transmitted and NACK received
// case TWI_MTX_ADR_NACK:                      	   // SLA+W has been transmitted and NACK received
// case TWI_MRX_ADR_NACK:                      	   // SLA+R has been transmitted and NACK received
// case TWI_NO_STATE;                         	   // No relevant state information available; TWINT = 0
// case TWI_BUS_ERROR:                        	   // Bus error due to an illegal START or STOP condition

       default:                                    // Reset TWI Interface
         TWCR =(1<<TWEN)|                         	// Enable TWI-interface and release TWI pins
               (0<<TWIE)|(1<<TWINT)|              	// Disable Interupt, clear Interrupt Flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// No Signal requests - nderung: Stop Condition senden
               (0<<TWWC);
         break;
      }                                            // switch
   }                                               // (TWSR > 0x5f)
}

// --- [ eof ] ----------------------------------------------------------------
